@@ -261,7 +261,10 @@
   ;; if the defstruct was compiled in a policy of SPEED 3.
   (equalp-impl #'equalp-err :type (sfunction (t t) boolean) :read-only t)
   ;; Information for the quicker variant of SLOT-VALUE on STRUCTURE-OBJECT
-  (struct-slot-map nil)
+  ;; INSTANCE uses at most 14 bits in the primitive object header for the payload
+  ;; length, so the function can't actually return all of the INDEX type.
+  (slot-mapper nil :type (or (sfunction (symbol) (or index null))
+                             simple-vector null))
   ;; Information about slots in the class to PCL: this provides fast
   ;; access to slot-definitions and locations by name, etc.
   ;; See MAKE-SLOT-TABLE in pcl/slots-boot.lisp for further details.
@@ -566,6 +569,9 @@
       (member-type
        (!alloc-member-type bits (member-type-xset x) (member-type-fp-zeroes x))))))
 #-sb-xc-host
+(macrolet ((safe-member-type-elt-p (obj)
+             `(or (not (sb-vm:is-lisp-pointer (get-lisp-obj-address ,obj)))
+                  (heap-allocated-p ,obj))))
 (defun copy-ctype (x &optional (flags 0))
   (declare (type ctype x))
   (declare (sb-c::tlab :system) (inline !new-xset))
@@ -591,7 +597,7 @@
              ;; which is not dealt with here (hash-table in the arena)
              (cond ((listp data)
                     ;; the XSET can be empty if a MEMBER type contains only FP zeros.
-                    ;; While we could use (load-time-value) to referece a constant empty xset
+                    ;; While we could use (load-time-value) to reference a constant empty xset
                     ;; there's really no point to doing that.
                     (collect ((elts))
                       (dolist (x data (!new-xset (elts) (xset-extra xset)))
@@ -651,6 +657,7 @@
         #+sb-simd-pack-256
         (simd-pack-256-type (!alloc-simd-pack-256-type bits (simd-pack-256-type-tag-mask x)))
         (alien-type-type (!alloc-alien-type-type bits (alien-type-type-alien-type x)))))))
+) ; end  MACROLET
 
 #-sb-xc-host
 (progn
