@@ -236,7 +236,7 @@
            (end-addr (ea (sb-x86-64-asm::ea-segment free-pointer)
                          (+ n-word-bytes (ea-disp free-pointer))
                          (ea-base free-pointer))))
-      (cond ((typep size `(integer ,large-object-size))
+      (cond ((and (integerp size) (>= size (target-heap-large-object-size)))
              ;; large objects will never be made in a per-thread region
              (cond (overflow (funcall overflow))
                    (t (fallback size)
@@ -408,7 +408,7 @@
        (inst lea result (ea list-pointer-lowtag rsp-tn)))
       (t
        (let ((nbytes (* cons-size n-word-bytes))
-             (zeroed #+mark-region-gc t)
+             (zeroed (target-heap-prezeroed-p))
              (prev-constant temp)) ;; a non-eq initial value
          (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn)
          (pseudo-atomic (:thread-tn thread-tn)
@@ -432,7 +432,7 @@
   (:policy :fast-safe)
   (:generator 10
     (let ((nbytes (* cons-size 2 n-word-bytes))
-          (zeroed #+mark-region-gc t)
+          (zeroed (target-heap-prezeroed-p))
           (prev-constant temp))
       (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn)
       (pseudo-atomic (:thread-tn thread-tn)
@@ -477,7 +477,7 @@
        (inst lea result (ea list-pointer-lowtag rsp-tn)))
       (t
        (let ((nbytes (* cons-size 2 n-word-bytes))
-             (zeroed #+mark-region-gc t)
+             (zeroed (target-heap-prezeroed-p))
              (prev-constant temp))
          (instrument-alloc +cons-primtype+ nbytes node (list temp alloc) thread-tn)
          (pseudo-atomic (:thread-tn thread-tn)
@@ -503,7 +503,7 @@
     (aver (>= cons-cells 3)) ; prevent regressions in ir2tran's vop selection
     (let* ((stack-allocate-p (node-stack-allocate-p node))
            (size (* (pad-data-block cons-size) cons-cells))
-           (zeroed #+mark-region-gc (not stack-allocate-p))
+           (zeroed (and (not stack-allocate-p) (target-heap-prezeroed-p)))
            (prev-constant temp))
       (unless stack-allocate-p
         (instrument-alloc +cons-primtype+ size node (list ptr temp) thread-tn))
@@ -984,8 +984,7 @@
                           (inst or temp layout)))
                    temp)
                  result 0 fun-pointer-lowtag (not stack-allocate-p))
-        (inst lea (pc-size vop)
-              temp (rip-relative-ea label (ash simple-fun-insts-offset word-shift)))
+        (inst lea temp (rip-relative-ea label (ash simple-fun-insts-offset word-shift)))
         (storew temp result closure-fun-slot fun-pointer-lowtag)))))
 
 (define-vop (reference-closure)
@@ -993,7 +992,7 @@
   (:results (result :scs (descriptor-reg)))
   (:vop-var vop)
   (:generator 1
-    (inst lea (pc-size vop) result (rip-relative-ea label fun-pointer-lowtag))))
+    (inst lea result (rip-relative-ea label fun-pointer-lowtag))))
 
 ;;; The compiler likes to be able to directly make value cells.
 (define-vop (make-value-cell)
@@ -1160,12 +1159,8 @@
       (move c-arg-1 rax-tn))
     (move res c-arg-1)))
 
-#+immobile-space
-(macrolet ((c-fun (name)
-             `(let ((c-fun (make-fixup ,name :foreign)))
-                (cond ((sb-c::code-immobile-p node) c-fun)
-                                 (t (progn (inst mov rax c-fun) rax))))))
-(define-vop (!alloc-immobile-fixedobj)
+#+(and sb-xc-host immobile-space)
+(define-vop (alloc-immobile-fixedobj)
   (:args (size-class :scs (any-reg) :target c-arg1)
          (nwords :scs (any-reg) :target c-arg2)
          (header :scs (any-reg) :target c-arg3))
@@ -1177,7 +1172,6 @@
                :offset #.(third *c-call-register-arg-offsets*)) c-arg3)
   (:temporary (:sc unsigned-reg :from :eval :to (:result 0) :offset rax-offset) rax)
   (:results (result :scs (descriptor-reg)))
-  (:node-var node)
   (:generator 50
    (inst mov c-arg1 size-class)
    (inst mov c-arg2 nwords)
@@ -1186,7 +1180,7 @@
    ;; which has that effect
    (inst and rsp-tn -16)
    (pseudo-atomic ()
-     (call-c (c-fun "alloc_immobile_fixedobj"))
+     (call-c "alloc_immobile_fixedobj")
      (move result rax))))
 
 ;;; Timing test:
@@ -1256,5 +1250,3 @@
        FAIL
        (inst mov result nil-value)
        OUT)))
-
-) ; end MACROLET
